//----------------------------------------------------------------------------------------
// mesh_2d_sw.H: Contains declarations for 2D node, edge, and triangle shallow-water objects
//----------------------------------------------------------------------------------------

#ifndef MESH_2D_SW_H
#define MESH_2D_SW_H

#include <cassert>
#include <fstream>
#include "general.H"
#include "map_info.H"
#include "mesh_2d.H"
#include "quadrature.H"
#include "dof.H"
#include "basis.H"
#include "storage_container.H"

//--------------------------------------------------------------------------------------
// NODE_2D_SW Class: Node class for shallow-water problems
//--------------------------------------------------------------------------------------

class NODE_2D_SW : public NODE_2D
{
private:
  real xi, depth;                // Surface elevation and bathymetry at the node
  real coord_spher[d_space_2d];  // Spherical coordianates of the node

public:
  // Default constructor
  NODE_2D_SW();

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  NODE_2D_SW(std::ifstream& in);

  // Default destructor
  ~NODE_2D_SW() {}

  // Return value: spherical coordinates of the node
  const real* SpherCoord() const {return coord_spher;}

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate (spherical), rvalue
  real SpherCoord(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return coord_spher[i];
  }

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate (spherical), lvalue
  real& SpherCoord(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return coord_spher[i];
  }

  // Return value: bathymetric depth at the node, rvalue
  real Depth() const {return depth;}

  // Return value: bathymetric depth at the node, lvalue
  real& Depth() {return depth;}

  // Return value: surface elevation at the node, rvalue
  real Xi() const {return xi;}

  // Return value: surface elevation at the node, lvalue
  real& Xi() {return xi;}

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const NODE_2D_SW& node);
};



//--------------------------------------------------------------------------------------
// EDGE_2D_SW Class: Edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Data shared by all edges
struct tmp_data_edge
{
  // Pointers to storage for degrees of freedom
  DOF_iterator *Var_dof[d_var], *L_Var_dof[d_var];

  // Default constructor
  tmp_data_edge()
  {
    for (int i=0; i<d_var; i++) Var_dof[i]=L_Var_dof[i]=0;
  }
};

// Local data at each quadrature point on edge
struct quad_loc_data_edge
{
  // Data stored at the quadrature point
  real Var[2][d_var_2d_dt], depth, surface_elev, coord[d_space_2d];
  // Index to storage for basis functions at the quadrature point
  std::size_t basis_func_ind[2][d_var_2d];

  // Default constructor
  quad_loc_data_edge()
  {
    depth=surface_elev=coord[0]=coord[1]=0.;
    for (int i=0; i<2; i++)
    {
      for (int j=0; j<d_var_2d; basis_func_ind[i][j]=(std::size_t)BAD_INDEX, j++);
      for (int j=0; j<d_var_2d_dt; Var[i][j]=0., j++);
    }
  }
};

typedef StorageContainer<tmp_data_edge, empty, quad_loc_data_edge> BaseStorageEdgeSW;

class EDGE_2D_SW : public EDGE_2D, public BaseStorageEdgeSW
{
private:
  real xi_0, d_xi;       // Surface elevation computation coefficients xi=xi_0+xi*d_xi
  real dp_0, d_dp;       // Depth computation coefficients depth=dp_0+xi*d_dp
  QUADRATURE *quad;      // Pointer to the quadrature object corresponding to the edge
  long_real coef_scale_d[2];  // Scaling coefficients for quadrature integration

protected:
  char n_basis[2];            // Number of different basis functions
  char basis_index[2][d_var]; // Index of basis functions corresponding to different variables

public:
  enum edge_type {UNKNOWN, INTERIOR, LAND, SEA, RIVER, RADIATION};

  // Default constructor
  EDGE_2D_SW();

  // Constructor that copies data from an object of EDGE_2D type
  // Parameters: edge - object to be copied
  EDGE_2D_SW(const EDGE_2D& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_2D_SW(std::ifstream& in);

  // Default destructor
  ~EDGE_2D_SW() {}

  // Compute mapping from the reference edge
  void ComputeReferenceMapping();

  // Return value: pointer to edge of the same type as the caller
  virtual EDGE_2D_SW* CreateEdge()
  {EDGE_2D_SW* edge = new EDGE_2D_SW(); assert(edge!=0); return edge;}

  // Return value: pointer to quadrature object for the edge
  QUADRATURE* Quad() const {return quad;}

  // Calculate coefficients for surface elevations computation
  void ComputeSurfaceElevation();

  // Calculate coefficients for surface elevations computation
  // Parameters: elev - surface elevation at the nodes
  void ComputeSurfaceElevation(real elev[2]);

  // Parameters: xi - point coordinates
  // Return value: surface elevation at xi
  real SurfaceElevation(const real xi[]) const {return xi_0+xi[0]*d_xi;}

  // Parameters: xi - point coordinates
  // Return value: depth at xi
  real Depth(const real xi[]) const {return dp_0+xi[0]*d_dp;}

  // Update recursively edge-element connectivity info
  // Parameters: el_ind - element index in the edge; element_new - element id
  void SetElementRecursive(int el_ind, int element_new);

  // Set recursively the edge (or its descendants) to be active
  // This version works for bundary edges and must be overloaded for interior edges
  virtual void SetActiveRecursive();

  // Set recursively the edge (or its descendants) to be inactive
  void UnSetActiveRecursive();

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the edge using bisection
  // Parameters: node_new_ind - id of the new node, lvalue
  // Return value: id of the first child edge
  virtual void Refine(int& node_new_ind);
#endif

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, rvalue
  long_real coef_scale(int el_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, lvalue
  long_real& coef_scale(int el_ind)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Parameters: el_ind - element index on the edge; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" for unknown "var_ind"
  DOF_iterator& Var_dof(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && 
                 var_ind<d_var && VAR_2D_INDEX_INV[var_ind]!=BAD_INDEX);
    temp().Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the edge; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" 
  //               for right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof(int el_ind,  VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && 
                 var_ind<d_var && VAR_2D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    temp().L_Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().L_Var_dof[var_ind];
  }

  // ----------- Accessors: Deal with data stored at quadrature points -------------------

  // Parameters: el_ind - element index on the edge; var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, rvalue
  const real* basis_func_quad(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && 
                 var_ind<d_var && basis_index[el_ind][var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_func_ind[el_ind][basis_index[el_ind][var_ind]]);
  }

  // Parameters: el_ind - element index on the edge; i - index of the basis function
  // Return value: pointer to values of basis functions at the quadrature point, lvalue
  real* basis_func_assign_quad(int el_ind, int i)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && i>=0 && i<n_basis[el_ind]);
    return &mass_storage(quad_local().basis_func_ind[el_ind][i]);
  }

  // Return value: depth at the quadrature point, rvalue
  real depth_quad() const {return quad_local().depth;}

  // Return value: depth at the quadrature point, lvalue
  real& depth_quad() {return quad_local().depth;}

  // Return value: surface elevation at the quadrature point, rvalue
  real surface_elev_quad() const {return quad_local().surface_elev;}

  // Return value: surface elevation at the quadrature point, lvalue
  real& surface_elev_quad() {return quad_local().surface_elev;}

  // Return value: coordinates of the quadrature point
  real *coord_quad() {return quad_local().coord;}

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, rvalue
  real coord_quad(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return quad_local().coord[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, lvalue
  real& coord_quad(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return quad_local().coord[i];
  }

  // Parameters: el_ind - element index on the edge; var_ind - global index of the unknown
  // Return value: value of the unknown at the quadrature point, rvalue
  real Var_quad(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && var_ind>=0 && var_ind<d_var &&
                 VAR_2D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    return quad_local().Var[el_ind][VAR_2D_DT_INDEX_INV[var_ind]];
  }

#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  // Parameters: i - coordinate index
  // Return value: value of the depth integrated velocity at the quadrature point, rvalue
  real UH_quad(int i)
  {
    static VAR ind[d_space_2d]={UH, VH};
    assert_debug(i>=0 && i<d_space_2d);
    if (IsElement(1))
      return 0.5*(Var_dof(0, ind[i])*basis_func_quad(0, ind[i])+Var_dof(1, ind[i])*basis_func_quad(1, ind[i]));
    else
      return Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
  }
#endif

// #ifdef BAROCLINIC_KEY
//   // Return value: depth integrated density forcing term at the quadrature point
//   //               works for most boundary edge types
//   virtual long_real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0));
//     return Var_dof(0, RO_FORCING)*basis_func_quad(0, RO_FORCING);
//   }
// #endif

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  virtual void BC_Dirichlet_quad(real t, const real param[], real var[d_var])
  {std::cerr << "File mesh_3d_sw.H: Dirichlet boundary condition is not spacified for this face type!\n";
    std::exit(1);}

  //----------------- Updaters: Update data stored at quadrature points ------------------

  // Initialize storage for data stored at quadrature points
  virtual void InitializeStorage();

  // Update data shared by all edges
  // Parameters: var_new - global storage for degrees of freedom;
  //             var_rhs_new - global storage for degrees of freedom for the right hand side
  static void Update_static(DOF var_new[d_var], DOF var_rhs_new[d_var]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  virtual void Update_basis_quad();

  // Update often changed data at quadrature points (unknowns)
  // Parameters: t - time
  void Update_variables_quad(real t);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const EDGE_2D_SW&);
};



//--------------------------------------------------------------------------------------
// EDGE_INTERIOR Class: Interior edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_INTERIOR : public EDGE_2D_SW
{
public:

  // Default constructor
  EDGE_INTERIOR();

  // Constructor
  EDGE_INTERIOR(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_INTERIOR(std::ifstream& in) : EDGE_2D_SW(in) {}

  // Default destructor
  ~EDGE_INTERIOR() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_INTERIOR* CreateEdge()
  {EDGE_INTERIOR* edge = new EDGE_INTERIOR(); assert(edge!=0); return edge;}

  // Set recursively the edge (or its descendants) to be active
  void SetActiveRecursive();

// #ifdef BAROCLINIC_KEY
//   // Return value: depth integrated density forcing term at the quadrature point
//   long_real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0) && IsElement(1));
//     long_real ro_0=Var_dof(0, RO_FORCING)*basis_func_quad(0, RO_FORCING),
//       ro_1=Var_dof(1, RO_FORCING)*basis_func_quad(1, RO_FORCING);
//     return 0.5*(ro_0+ro_1);
//   }
// #endif

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);
};



//--------------------------------------------------------------------------------------
// EDGE_LAND Class: Land edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_LAND : public EDGE_2D_SW
{
public:
  // Default constructor
  EDGE_LAND();

  // Constructor
  EDGE_LAND(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_LAND(std::ifstream& in) : EDGE_2D_SW(in) {}

  // Default destructor
  ~EDGE_LAND() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_LAND* CreateEdge()
  {EDGE_LAND* edge = new EDGE_LAND(); assert(edge!=0); return edge;}

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);
};



//--------------------------------------------------------------------------------------
// EDGE_SEA Class: Sea edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_SEA : public EDGE_2D_SW
{
private:
  int n_freq;                              // Number of forcing frequencies in the edge
  std::size_t amplitude_ind;               // Relative amplitude on an open sea edge
  std::size_t phase_ind;                   // Phase on an open sea edge
  std::size_t tidal_coef_ind[d_space_2d];  // Coefficients for tidal forcing computation
  real tidal_elevation;                    // Tidal elevation at the edge
#ifdef BAROCLINIC_KEY
  real temperature, salinity;              // temperature and salinity at the edge
#endif

public:
  // Default constructor
  EDGE_SEA();

  // Constructor
  EDGE_SEA(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_SEA(std::ifstream& in);

  // Default destructor
  ~EDGE_SEA() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_SEA* CreateEdge()
  {EDGE_SEA* edge = new EDGE_SEA(); assert(edge!=0); return edge;}

  // Return value: number of forcing frequencies, rvalue
  int NFreq() const {return n_freq;}

  // Return value: number of forcing frequencies, rvalue
  int& NFreq() {return n_freq;}

  // Return value: pointer to array with values of forcing amplitude
  const real* Amplitude() const {return &mass_storage(amplitude_ind);}

  // Parameters: i - coordinate index
  // Return value: amplitude for the i-th component, rvalue 
  real Amplitude(int i) const
  {
    assert_debug(i>=0 && i<n_freq);
    return mass_storage(amplitude_ind+i);
  }

  // Parameters: i - coordinate index
  // Return value: amplitude for the i-th component, lvalue 
  real& Amplitude(int i)
  {
    assert_debug(i>=0 && i<n_freq);
    return mass_storage(amplitude_ind+i);
  }

  // Return value: pointer to array with values of forcing phase
  const real* Phase() const {return &mass_storage(phase_ind);}

  // Parameters: i - coordinate index
  // Return value: phase for the i-th component, rvalue 
  real Phase(int i) const
  {
    assert_debug(i>=0 && i<n_freq);
    return mass_storage(phase_ind+i);
  }

  // Parameters: i - coordinate index
  // Return value: phase for the i-th component, lvalue 
  real& Phase(int i)
  {
    assert_debug(i>=0 && i<n_freq);
    return mass_storage(phase_ind+i);
  }

  // Parameters: i - forcing component index; d - coordinate index
  // Return value: coefficient for tidal forcing computation, rvalue 
  real TidalCoef(int i, int d) const
  {
    assert_debug(i>=0 && i<n_freq && d>=0 && d<d_space_2d);
    return mass_storage(tidal_coef_ind[d]+i);
  }

  // Parameters: i - forcing component index; d - coordinate index
  // Return value: coefficient for tidal forcing computation, lvalue 
  real& TidalCoef(int i, int d)
  {
    assert_debug(i>=0 && i<n_freq && d>=0 && d<d_space_2d);
    return mass_storage(tidal_coef_ind[d]+i);
  }

  // Compute coefficients for tidal forcing
  // Parameters: i - forcing component index
  void ComputeTidalCoefficients(int i);

  // Return value: tidal elevation on the edge, rvalue 
  real TidalElevation() const {return tidal_elevation;}

  // Return value: tidal elevation on the edge, lvalue 
  real& TidalElevation() {return tidal_elevation;}

#ifdef BAROCLINIC_KEY
  // Return value: temperature on the edge, rvalue 
  real Temperature() const {return temperature;}

  // Return value: temperature on the edge, lvalue 
  real& Temperature() {return temperature;}

  // Return value: salinity on the edge, rvalue 
  real Salinity() const {return salinity;}

  // Return value: salinity on the edge, lvalue 
  real& Salinity() {return salinity;}
#endif

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the edge using edge bisection
  // Parameters: node_new_ind - id of the new node, lvalue
  // Return value: id of the first child edge
  void Refine(int& node_new_ind);
#endif

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const EDGE_SEA&);
};



//--------------------------------------------------------------------------------------
// EDGE_RIVER Class: River edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_RIVER : public EDGE_2D_SW
{
private:
  real elevation;             // Surface elevation from the side of river
  real velocity[d_space_2d];  // Inflow velocity
#ifdef BAROCLINIC_KEY
  real temperature, salinity; // Temperature and salinity of the inflow
#endif

public:
  // Default constructor
  EDGE_RIVER();

  // Constructor
  EDGE_RIVER(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_RIVER(std::ifstream& in);

  // Default destructor
  ~EDGE_RIVER() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_RIVER* CreateEdge()
  {EDGE_RIVER* edge = new EDGE_RIVER(); assert(edge!=0); return edge;}

  // Return value: surface elevation on the river edge, rvalue
  real SurfaceElevation() const {return elevation;}

  // Return value: surface elevation on the river edge, lvalue
  real& SurfaceElevation() {return elevation;}

  // Parameters: i - coordinate index
  // Return value: inflow velocity on the river edge, rvalue
  real Velocity(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return velocity[i];
  }

  // Parameters: i - coordinate index
  // Return value: inflow velocity on the river edge, lvalue
  real& Velocity(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return velocity[i];
  }

#ifdef BAROCLINIC_KEY
  // Return value: inflow temperature on the river edge, rvalue
  real Temperature() const {return temperature;}

  // Return value: inflow temperature on the river edge, lvalue
  real& Temperature() {return temperature;}

  // Return value: inflow salinity on the river edge, rvalue
  real Salinity() const {return salinity;}

  // Return value: inflow salinity on the river edge, lvalue
  real& Salinity() {return salinity;}
#endif

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the edge using edge bisection
  // Parameters: node_new_ind - id of the new node, lvalue
  // Return value: id of the first child edge
  void Refine(int& node_new_ind);
#endif

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const EDGE_RIVER&);
};



//--------------------------------------------------------------------------------------
// EDGE_RADIATION Class: Radiation edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_RADIATION : public EDGE_2D_SW
{
public:
  // Default constructor
  EDGE_RADIATION();

  // Constructor
  EDGE_RADIATION(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_RADIATION(std::ifstream& in) : EDGE_2D_SW(in) {}

  // Default destructor
  ~EDGE_RADIATION() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_RADIATION* CreateEdge()
  {EDGE_RADIATION* edge = new EDGE_RADIATION(); assert(edge!=0); return edge;}

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);
};


#ifdef TEST_KEY
//--------------------------------------------------------------------------------------
// EDGE_UNKNOWN Class: Test (Dirichlet) edge class for shallow-water problems
//--------------------------------------------------------------------------------------

class EDGE_UNKNOWN : public EDGE_2D_SW
{
public:
  // Default constructor
  EDGE_UNKNOWN();

  // Constructor
  EDGE_UNKNOWN(const EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  EDGE_UNKNOWN(std::ifstream& in) : EDGE_2D_SW(in) {}

  // Default destructor
  ~EDGE_UNKNOWN() {}

  // Return value: pointer to edge of the same type as the caller
  EDGE_UNKNOWN* CreateEdge()
  {EDGE_UNKNOWN* edge = new EDGE_UNKNOWN(); assert(edge!=0); return edge;}

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);
};
#endif



//--------------------------------------------------------------------------------------
// TRIANGLE_2D_SW Class: Element class in 2D for shallow-water problems
//--------------------------------------------------------------------------------------

// Data shared by all elements
struct tmp_data_triangle
{
  // Field used for norm and error computation
  long_real norm;
  // Pointers to storage for degrees of freedom
  DOF_iterator *Var_dof[d_var], *L_Var_dof[d_var];
  // Index to storage for basis functions at the vertices
  std::size_t basis_func_node_ind[d_var_2d][vert_h];

  // Default constructor
  tmp_data_triangle()
  {
    norm=0;
    for (int i=0; i<d_var_2d; i++) 
      for (int j=0; j<vert_h; basis_func_node_ind[i][j]=(std::size_t)BAD_INDEX, j++);
    for (int i=0; i<d_var; i++)
      Var_dof[i]=L_Var_dof[i]=0;
  }
};

// Data at quadrature points shared by all elements
struct quad_tmp_data_triangle
{
  long_real weight;  // Weight at the quadrature point
  // Index to storage for basis functions at the quadrature point
  std::size_t basis_func_ind[d_var_2d];

  // Default constructor
  quad_tmp_data_triangle()
  {
    weight=0.;
    for (int i=0; i<d_var_2d; basis_func_ind[i]=(std::size_t)BAD_INDEX, i++);
  }
}; 

// Local data at each quadrature point on element
struct quad_loc_data_triangle
{
  // Data stored at the quadrature point
  real Var[d_var_2d_dt], depth, surface_elev, Coriolis, coord[d_space_2d];
  // Index to storage for gradients of basis functions at the quadrature point
  std::size_t basis_grad_ind[d_var_2d][d_space_2d];

  // Default constructor
  quad_loc_data_triangle()
  {
    depth=surface_elev=Coriolis=coord[0]=coord[1]=0.;
    for (int i=0; i<d_var_2d_dt; Var[i]=0., i++);
    for (int i=0; i<d_var_2d; i++)
      for (int j=0; j<d_space_2d; basis_grad_ind[i][j]=(std::size_t)BAD_INDEX, j++);
  }
};

typedef StorageContainer<tmp_data_triangle, quad_tmp_data_triangle,
quad_loc_data_triangle> BaseStorageTriangleSW;

class TRIANGLE_2D_SW : public TRIANGLE_2D, public BaseStorageTriangleSW
{
private:
  unsigned char order;            // Order of the approximating space
  real xi_0, d_xi[d_space_2d];    // Coefficients for surface elevation computation
  real dp_0, d_dp[d_space_2d];    // Coefficients for depth computation
  std::size_t tidal_coef_ind;     // Coefficients for tidal forcing computation
  real tidal_forcing[d_space_2d]; // Tidal potential forcing term
  QUADRATURE *quad;               // Pointer to quadrature object corresponding to the element
  QUADRATURE *vert;               // Pointer to quadrature object for the vertices of the element
#ifdef ADAPTIVITY_H_2D_KEY
  real min, max, abs_max;
  real s_jump[2], s_boundary[2];  // Variables for shock-detector
#endif
protected:
  BASIS *basis[d_var_2d];         // Pointers to objects describing the basis functions
  char n_basis;                   // Number of different basis functions
  char basis_index[d_var];        // Index of basis functions corresponding to different variables

public:
  // Default constructor
  TRIANGLE_2D_SW();

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  TRIANGLE_2D_SW(std::ifstream& in);

  // Default destructor
  ~TRIANGLE_2D_SW() {}

  // Compute mapping from the reference triangle
  void ComputeReferenceMapping();

  // Return value: order of the approximation on the element, rvalue
  unsigned char Order() const {return order;}

  // Return value: order of the approximation on the element, lvalue
  unsigned char& Order() {return order;}

  // Return value: pointer to quadrature object for the element
  QUADRATURE* Quad() const {return quad;}

  // Return value: pointer to quadrature object for the vertices
  QUADRATURE* Vert() const {return vert;}

  // Parameters: i - basis index
  // Return value: pointer to basis object, rvalue
  const BASIS *Basis(int i) const
  {
    assert_debug(i>=0 && i<d_var_2d);
    return basis[i];
  }

  // Parameters: var_ind - variable index
  // Return value: pointer to basis object, rvalue
  const BASIS *Basis_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && basis_index[var_ind]!=BAD_INDEX);
    return basis[basis_index[var_ind]];
  }

  // Return value: number of different basis functions
  char NBasis() const {return n_basis;}  

  // Parameters: var_ind - variable index
  // Return value: index of basis object corresponding to the unknown
  char Basis_index(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var);
    return basis_index[var_ind];
  }

  // Parameters: xi - point coordinates
  // Return value: surface elevation at xi
  real SurfaceElevation(const real xi[]) const 
  {return xi_0+xi[0]*d_xi[0]+xi[1]*d_xi[1];}

  // Calculate coefficients for surface elevations computation
  void ComputeSurfaceElevation();

  // Parameters: xi - point coordinates
  // Return value: depth at xi
  real Depth(const real xi[]) const 
  {return dp_0+xi[0]*d_dp[0]+xi[1]*d_dp[1];}

  // Calculate coefficients for depth computation
  void ComputeDepth(const real depth[])
  {dp_0=depth[0]; d_dp[0]=depth[1]-depth[0]; d_dp[1]=depth[2]-depth[0];}

  // Check minimum depth condition and adjust surface elevation if necessary
  // Parameters: min_depth - minimum depth
  void MinDepthCheck(real min_depth);

  // Parameters: i - forcing component index; a, b - coordinate indices
  // Return value: coefficient for tidal forcing computation, rvalue 
  real TidalCoef(int i, int a, int b) const {return mass_storage(tidal_coef_ind+4*i+2*a+b);}

  // Parameters: i - forcing component index; a, b - coordinate indices
  // Return value: coefficient for tidal forcing computation, rvalue 
  real& TidalCoef(int i, int a, int b) {return mass_storage(tidal_coef_ind+4*i+2*a+b);}

  // Parameters: i - coordinate index
  // Return value: tidal forcing term for the momentum equation, rvalue 
  real TidalForcing(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return tidal_forcing[i];
  }

  // Parameters: i - coordinate index
  // Return value: tidal forcing term for the momentum equation, lvalue 
  real& TidalForcing(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return tidal_forcing[i];
  }

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the element using edge bisection
  // Return value: id of the first child element
  virtual void Refine();

  // Unrefine the element
  virtual void UnRefine();
#endif

#ifdef ADAPTIVITY_H_2D_KEY
  // Initialize data used for shock-detector
  void InitializeAdaptivityData();

  // Return value: maximum value of the unknown used for adaptivity on the element
  real GetMax() const {return max;}

  // Return value: minimum value of the unknown used for adaptivity on the element
  real GetMin() const {return min;}

  // Return value: maximum absolute value of the unknown used for adaptivity on the element
  real GetMaxAbs() const {return abs_max;}

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, rvalue
  real JumpBoundary(int k) const
  {
    assert_debug(k>=0 && k<2);
    return s_boundary[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, lvalue
  real& JumpBoundary(int k)
  {
    assert_debug(k>=0 && k<2);
    return s_boundary[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, rvalue
  real Jump(int k) const
  {
    assert_debug(k>=0 && k<2);
    return s_jump[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, lvalue
  real& Jump(int k)
  {
    assert_debug(k>=0 && k<2);
    return s_jump[k];
  }
#endif

  // Parameters: var_ind - global index of the unknown, j - node index
  // Return value: pointer to values of basis functions at the node, rvalue
  const real* basis_func_node(VAR var_ind, int j) const 
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<vert_h &&
      basis_index[var_ind]!=BAD_INDEX);
    return &mass_storage(temp().basis_func_node_ind[basis_index[var_ind]][j]);
  }

  // Return value: scaling coefficient for quadrature integration, rvalue
  long_real scale() const {return Mapping().GetScale();}

  // Return value: value in the "norm" field, lvalue
  static long_real& norm() {return temp_storage.norm;}

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on element for unknown "var_ind"
  DOF_iterator& Var_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_2D_INDEX_INV[var_ind]!=BAD_INDEX);
    temp().Var_dof[var_ind]->SetElement(ID());
    return *temp().Var_dof[var_ind];
  }

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on element for right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_2D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    temp().L_Var_dof[var_ind]->SetElement(ID());
    return *temp().L_Var_dof[var_ind];
  }

  // --------------- Accessors: Deal with data stored at quadrature points ---------------

  // Coriolis coefficient at the quadrature point, rvalue
  real Coriolis_quad() const {return quad_local().Coriolis;}

  // Coriolis coefficient at the quadrature point, lvalue
  real& Coriolis_quad() {return quad_local().Coriolis;}

  // Weight at the quadrature point, rvalue
  long_real weight_quad() const {return quad_temp().weight;}

  // Parameters: i - index of the quadrature point
  // Weight at the quadrature point, lvalue
  static long_real& weight_quad(int i) {return quad_temp_storage[i].weight;}

  // Return value: coordinates of the quadrature point, lvalue
  const real *coord_quad() const {return quad_local().coord;}

  // Return value: coordinates of the quadrature point, lvalue
  real *coord_quad() {return quad_local().coord;}

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, rvalue
  real coord_quad(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return quad_local().coord[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, lvalue
  real& coord_quad(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return quad_local().coord[i];
  }

  // Return value: depth at the quadrature point, rvalue
  real depth_quad() const {return quad_local().depth;}

  // Return value: depth at the quadrature point, lvalue
  real& depth_quad() {return quad_local().depth;}

  // Return value: surface elevation at the quadrature point, rvalue
  real surface_elev_quad() const {return quad_local().surface_elev;}

  // Return value: surface elevation at the quadrature point, lvalue
  real& surface_elev_quad() {return quad_local().surface_elev;}

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, rvalue
  const real* basis_func_quad(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && basis_index[var_ind]!=BAD_INDEX);
    return &mass_storage(quad_temp().basis_func_ind[basis_index[var_ind]]);
  }

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of basis functions at the quadrature point, lvalue
  real* basis_func_assign_quad(int i)
  {
    assert_debug(i>=0 && i<n_basis);
    return &mass_storage(quad_temp().basis_func_ind[i]);
  }

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of basis functions at the quadrature point, rvalue
  const real* basis_grad_quad(VAR var_ind, int j) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space_2d &&
                 basis_index[var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_grad_ind[basis_index[var_ind]][j]);
  }

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of basis functions at the quadrature point, lvalue
  real* basis_grad_assign_quad(int i, int j)
  {
    assert_debug(i>=0 && i<n_basis && j>=0 && j<d_space_2d);
    return &mass_storage(quad_local().basis_grad_ind[i][j]);
  }

  // Parameters: var_ind - global index of the unknown
  // Return value: value of the unknown at the quadrature point, rvalue
  real Var_quad(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_2D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    return quad_local().Var[VAR_2D_DT_INDEX_INV[var_ind]];
  }

#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  // Parameters: i - coordinate index
  // Return value: value of the depth integrated velocity at the quadrature point, rvalue
  real UH_quad(int i)
  {
    static VAR ind[d_space_2d]={UH, VH};
    assert_debug(i>=0 && i<d_space_2d);
    return Var_dof(ind[i])*basis_func_quad(ind[i]);
  }
#endif

// #ifdef BAROCLINIC_KEY
//   // Return value: depth integrated density forcing term at the quadrature point
//   real density_forcing_quad() const {return Var_dof(RO_FORCING)*basis_func_quad(RO_FORCING);}
// #endif

  // ---------------- Updaters: Update data stored at quadrature points ------------------

  // Initialize storage for data stored at quadrature points
  virtual void InitializeStorage();

  // Update data shared by all edges
  // Parameters: quad - pointer to quadrature object;
  //             vert - pointer to quadrature object for vertices;
  //             basis - pointer to basis object
  //             var_new - global storage for degrees of freedom;
  //             var_rhs_new - global storage for degrees of freedom for the right hand side
  static void Update_static(const QUADRATURE *quad, const QUADRATURE *vert,
                            const BASIS *basis[d_var_2d], DOF var_new[d_var],
                            DOF var_rhs_new[d_var]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  virtual void Update_basis_quad();

  // Update often changed data at quadrature points (unknowns)
  // Parameters: t - time
  virtual void Update_variables_quad(real t);


  // Numerics: Functions performing quadrature integrations
  void Projection_elevation(real, void fn(real, const real[], real[]));
  void Projection_elevation_forcing(real, void fn(real, const real[], real[]));
  void Elevation_change_L2();
  void Elevation_error_L2(real, void fn(real, const real[], real[]));

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const TRIANGLE_2D_SW&);
};

#endif
