//----------------------------------------------------------------------------------------
// mesh_container.H: Contains definition of the mesh container
//----------------------------------------------------------------------------------------

#ifndef MESH_CONTAINER_H
#define MESH_CONTAINER_H

#include <iostream>
#include <cassert>
#include <deque>
#include "general.H"
#include "recording_stations.H"
#include "mesh_2d_sw.H"
#include "mesh_3d_sw.H"

//----------------------------------------------------------------------------------------
// MULTI_DEQUE template class: A list class containing objects of various sub-types
//                             derived from a common base class
//----------------------------------------------------------------------------------------

template<typename SType>
class MULTI_DEQUE
{
private:
  std::deque<SType*> index;  // Index-list (list of pointers to all entries in the subtype lists)

public:
  // Default constructor
  MULTI_DEQUE() {}

  // Default destructor
  ~MULTI_DEQUE() {}

  // Return value: size of the list
  std::size_t size() const {return index.size();}

  // Parameters: i - list index
  // Return value: i-th list entry, rvalue
  const SType& operator[](std::size_t i) const
  {
    assert_debug(i<index.size());
    return *index[i];
  }

  // Parameters: i - list indeex
  // Return value: i-th list entry, lvalue
  SType& operator[](std::size_t i)
  {
    assert_debug(i<index.size());
    return *index[i];
  }

  // Push back template function for MULTI_DEQUE class
  // Parameters: list - list of subtype objects, elem - a subtype entry to be added to the list
  template<typename T>
  void push_back_index(std::deque<T>& list, const SType& entry)
  {
    int id=entry.ID();
    std::size_t size=list.size();

    assert_debug(id!=BAD_INDEX);
    list.push_back((T&)entry);
    if (index.size()<=id)
      index.resize(id+1, 0); // Resize index-list if necessary
    index[id]=&list[size];
  }

  // Overloaded push_back function of std::deque
  // Parameters: elem - a subtype entry to be added to the list
  void push_back(const SType& entry);
};



//----------------------------------------------------------------------------------------
// MESH class: A list class for storage of mesh components
//----------------------------------------------------------------------------------------

class MESH
{
private:
  int status;                        // Mesh status
  int iteration_counter;             // Global iteration counter
  int update_counter;                // Mesh update counter
  int output_counter;                // Output update counter
  long int tecplot_output_position;  // Output position in the tecplot output file
  std::size_t n_prism_active;        // Number of active elements

public:
  enum status_value {UPDATED=1};

  // Lists of various mesh components
  std::deque<NODE_2D_SW> node_2d_list;
  std::deque<EDGE_INTERIOR> edge_interior_list;
  std::deque<EDGE_LAND> edge_land_list;
  std::deque<EDGE_RIVER> edge_river_list;
  std::deque<EDGE_SEA> edge_sea_list;
  std::deque<EDGE_RADIATION> edge_radiation_list;
  std::deque<NODE_3D_SW> node_3d_list;
  std::deque<TRIANGLE_2D_SW> triangle_2d_list;
  std::deque<FACE_HORIZ_SURFACE> face_horiz_surface_list;
  std::deque<FACE_HORIZ_SURFACE_STRESS> face_horiz_surface_stress_list;
  std::deque<FACE_HORIZ_INTERIOR> face_horiz_interior_list;
  std::deque<FACE_HORIZ_BOTTOM_NO_SLIP> face_horiz_bottom_no_slip_list;
  std::deque<FACE_HORIZ_BOTTOM_NO_DRAG> face_horiz_bottom_no_drag_list;
  std::deque<FACE_HORIZ_BOTTOM_QUAD_DRAG> face_horiz_bottom_quad_drag_list;
  std::deque<FACE_HORIZ_BOTTOM_LIN_DRAG> face_horiz_bottom_lin_drag_list;
  std::deque<FACE_VERT_INTERIOR> face_vert_interior_list;
  std::deque<FACE_VERT_LAND> face_vert_land_list;
  std::deque<FACE_VERT_SEA> face_vert_sea_list;
  std::deque<FACE_VERT_RIVER> face_vert_river_list;
  std::deque<FACE_VERT_RADIATION> face_vert_radiation_list;
  std::deque<PRISM_PARALLEL> prism_parallel_list;
  std::deque<PRISM_NONPARALLEL> prism_nonparallel_list;
#ifdef TEST_KEY
  std::deque<EDGE_UNKNOWN> edge_unknown_list;
  std::deque<FACE_HORIZ_BOTTOM_TEST> face_horiz_bottom_test_list;
  std::deque<FACE_VERT_UNKNOWN> face_vert_unknown_list;
#endif
  // Lists containing multiple sub-types of objects
  MULTI_DEQUE<EDGE_2D_SW> edge_2d_list;
  MULTI_DEQUE<FACE_HORIZ_SW> face_horiz_list;
  MULTI_DEQUE<FACE_VERT_SW> face_vert_list;
  MULTI_DEQUE<PRISM_SW> prism_list;

  // Elevation recording stations
  RecordingStations<1, TRIANGLE_2D_SW, std::deque<TRIANGLE_2D_SW> > elev_stations;
  // Velocity recording stations
  RecordingStations<d_space_2d, TRIANGLE_2D_SW, std::deque<TRIANGLE_2D_SW> > vel_stations;

  // Default constructor
  MESH() {status=iteration_counter=0; update_counter=1; output_counter=9; tecplot_output_position=0;}

  // Default destructor
  ~MESH() {}

  // Input 2D mesh from a file (ADCIRC format)
  // Parameters: grid_coord - coordinates file (fort.14); grid_connect - connectivity file (fort.17);
  //             max_depth - maximum admissible bathymetric depth
  void Input2DMesh(const char* grid_coord, const char* grid_connect, real& max_depth);

  // Compute tidal forcing on elements and open sea edges
  // Parameters: t - time
  void ComputeTidalForcing(real t);

  // Create a vertical face of the type corresponding to the edge type
  // Parameters: edge - 2D edge corresponding to the vertical face
  FACE_VERT_SW* CreateFaceVert(EDGE_2D_SW& edge)
  {
    FACE_VERT_SW* face_vert;

    switch(edge.Type())
      {
      case EDGE_2D_SW::INTERIOR: face_vert = new FACE_VERT_INTERIOR((EDGE_INTERIOR&)edge);
        break;
      case EDGE_2D_SW::LAND: face_vert = new FACE_VERT_LAND((EDGE_LAND&)edge);
        break;
      case EDGE_2D_SW::SEA: face_vert = new FACE_VERT_SEA((EDGE_SEA&)edge);
        break;
      case EDGE_2D_SW::RIVER: face_vert = new FACE_VERT_RIVER((EDGE_RIVER&)edge);
        break;
      case EDGE_2D_SW::RADIATION: face_vert = new FACE_VERT_RADIATION((EDGE_RADIATION&)edge);
        break;
#ifdef TEST_KEY
      case EDGE_2D_SW::UNKNOWN: face_vert = new FACE_VERT_UNKNOWN((EDGE_UNKNOWN&)edge);
        break;
#endif
      default: std::cerr << "File mesh_container.H: Unknown edge type!\n";
        std::exit(1);
      }
    assert(face_vert!=0);
    return face_vert;
  }

  // Create 3D mesh
  // Parameters: max_depth - maximum admissible bathymetric depth
  void Create3DMesh(real max_depth);

  // Update free surface elevation
  // Parameters: t - time
  void UpdateSurfaceElevation(real t);

  // Update mesh in case of adaptive mesh refinement
  void UpdateMesh();

  // Return value: Has the mesh been updated?
  bool IsUpdated() const {return (status&UPDATED)!=0;};

  // Output in various formats
  // Parameters: t - time; output_ind - type of required output
  void Output(real t, int output_ind);

  // Project values of 3D unknowns to nodes for Tecplot output
  // Parameters: n_nodes - number of nodes, lvalue; n_elements - number of elements, lvalue;
  void ProjectToNodes(int& n_nodes, int& n_elements);

  // Output in Tecplot format
  // Parameters: t - time
  void TecplotOutput(real t);

  // Output in Tecplot format (for conforming grids)
  // Parameters: t - time
  void TecplotOutputConforming(real t);

  // Hot start output
  // Parameters: t - time
  void HotOutput(long_real t);

  // Hot start input
  // Parameters: t - time, lvalue
  void HotInput(long_real& t);

  // Return value: iteration counter, rvalue
  int IterationCounter() const {return iteration_counter;};

  // Return value: iteration counter, lvalue
  int& IterationCounter() {return iteration_counter;};

  // Return value: update counter, rvalue
  int UpdateCounter() const {return update_counter;};

  // Return value: update counter, lvalue
  int& UpdateCounter() {return update_counter;};

  // Increment active element counter
  void IncrementActiveElementCnt() {n_prism_active++;}

  // Decrement active element counter
  void DecrementActiveElementCnt() {n_prism_active--;}
};

#endif
